from pydantic import BaseModel
from pydantic_settings import BaseSettings
from typing import List

class Colors(BaseModel):
    base_colors: List[str] = ['#17A2B8', "#28A745", "#0D6EFD", "#DC3545", "#FFC107"]

class Settings(BaseSettings):
    colors: Colors = Colors()
    


from .report import get_report
from .empty_report import get_empty_report
from .comparison_report import get_comparison_report
from .error_report import get_error_report
from .test_report import get_test_report


from xurpas_data_quality.data.typeset import Numeric
from xurpas_data_quality.render.handler import Handler
from xurpas_data_quality.render.renderer import HTMLBase, HTMLContainer, HTMLTable, HTMLVariable, HTMLPlot, HTMLToggle, HTMLCollapse, HTMLDropdown
from xurpas_data_quality.visuals import plot_to_base64, create_tiny_histogram, create_histogram, create_distribution_plot, create_heatmap, create_interaction_plot


@Handler.register(Numeric)
def render_bottom_numerical(data, col_name:str,*args, **kwargs):
    quantile_statistics = {
        "Minimum": data['quantile_stats']['minimum'],
        "Fifth (5th) Percentile": data['quantile_stats']['5th_percentile'],
        "First Quartile (Q1)":data['quantile_stats']['Q1'],
        "Median": data['quantile_stats']['median'],
        "Ninety-Fifth (95th) Percentile": data['quantile_stats']["95th_percentile"],
        "Maximum": data['quantile_stats']['maximum'],
        "Range": data['quantile_stats']['range'],
        "Interquartile Range (IQR)": data['quantile_stats']['IQR']
    }

    descriptive_statistics = {
        "Standard Deviation": "{:.2f}".format(data['descriptive_stats']['std_dev']),
        "Mean": "{:.2f}".format(data['descriptive_stats']["mean"]),
        "Coefficient of Variation (CV)": "{:.2f}".format(data['descriptive_stats']["CV"]),
        "Kurtosis": "{:.2f}".format(data['descriptive_stats']['kurtosis']),
        "Mean Absolute Deviation": "{:.2f}".format(data['descriptive_stats']["MAD"]),
        "Skew": "{:.2f}".format(data['descriptive_stats']['skew']),
        "Sum": "{:.2f}".format(data['descriptive_stats']['sum']),
        "Variance": "{:.2f}".format(data['descriptive_stats']['variance']),
        "Monotonicity": data['descriptive_stats']["monotonicity"]
    }


    variable_bottom = [
        HTMLContainer(
            type="box",
            name="Overview",
            container_items=[HTMLContainer(
                type="default",
                name="Statistics",
                id="stats",
                container_items=[
                    HTMLContainer(
                        type = "column",
                        container_items = HTMLTable(
                            data = quantile_statistics,
                            name = 'Quantile Statistics')
                    ),
                    HTMLContainer(
                        type = "column",
                        container_items = HTMLTable(
                            data = descriptive_statistics,
                            name = 'Descriptive Statistics'
                        )
                    )
                ]
            )]
        ),
        HTMLPlot(
            name="Histogram",
            type="large",
            id="histo",
            plot=plot_to_base64(create_histogram(data['histogram']))
        )
    ]

    return HTMLContainer(
        type="tabs",
        col=col_name,
        container_items=variable_bottom
    )


import pandas as pd

from xurpas_data_quality.data.descriptions import TableDescription
from xurpas_data_quality.render.renderer import HTMLBase, HTMLContainer, HTMLTable
from xurpas_data_quality.config import Settings

def render_compare(data: TableDescription, name:str):
    
    config = {'color_list': Settings().colors.base_colors}

    headers = [f"Table {i+1}" for i in range(len(data))]
    distinct_count_list = []
    value_count_list = []
    shared_values_list = []
    variable_types_list = []
    num_variables_list = []
    missing_cells_list = []
    missing_cells_perc_list = []
    duplicate_rows_list = []
    duplicate_rows_perc_list = []

    for datum in data:
        distinct_count_list.append(datum.comparison['distinct_count'])
        value_count_list.append(datum.comparison['value_count'])
        shared_values_list.append(datum.shared_values)
        variable_types_list.append(datum.variable_types)
        num_variables_list.append(datum.dataset_statistics['num_variables'])
        missing_cells_list.append(datum.dataset_statistics['missing_cells'])
        missing_cells_perc_list.append("{:0.2f}%".format(datum.dataset_statistics['missing_cells_perc']))
        duplicate_rows_list.append(datum.dataset_statistics['duplicate_rows'])
        duplicate_rows_perc_list.append("{:0.2f}%".format(datum.dataset_statistics['duplicate_rows_perc']))

    combined_data = {
        'Distinct Values (Count)': distinct_count_list,
        'Values (Count)': value_count_list,
        'Values Shared Between Tables': shared_values_list
    }

    dataset_statistics = {
        'Number of Variables': num_variables_list,
        'Missing Cells': missing_cells_list,
        'Missing Cells (%)': missing_cells_perc_list,
        'Duplicate Rows': duplicate_rows_list,
        'Duplicate Rows (%)': duplicate_rows_perc_list 
    }

    variable_types_statistics = {}
    for key in set().union(*variable_types_list):
        variable_types_statistics[key] = [0] * len(variable_types_list)

    for i, d in enumerate(variable_types_list):
        for key, value in d.items():
            variable_types_statistics[key][i] = value

    column_comparison = HTMLContainer(
            type="box",
            name = "Comparison",
            container_items=[
                HTMLTable(
                    id = "comparison",
                    name = 'Comparing the "Volume" Column',
                    data = combined_data,
                    headers = headers,
                    config = config
                )
            ]
        )
    
    overview_section = HTMLContainer(
        type="box",
        name="Overview",
        container_items = [
            HTMLContainer(
                type="column",
                container_items = HTMLTable(
                    id = "comparison",
                    name = "Overview of Data",
                    data = dataset_statistics,
                    headers = headers,
                    config = config
                )
            ),
            HTMLContainer(
                type="column",
                container_items = HTMLTable(
                    id = "comparison",
                    name = "Overview of Variables",
                    data = variable_types_statistics,
                    headers = headers,
                    config = config
                )
            )
        ]
    )

    content = [
        column_comparison,
        overview_section
    ]
    body = HTMLContainer(type="sections",
                         container_items = content)

    if name is not None:
        return HTMLBase(
            body=body,
            name=name
        )
    
    else:
        return HTMLBase(
            body=body
        )


from xurpas_data_quality.data.descriptions import TableDescription
from xurpas_data_quality.render.render import render_report
from xurpas_data_quality.render.renderer import HTMLContainer, HTMLBase

from dataclasses import fields


def get_report(data: TableDescription,minimal:bool, name:str=None):
    """
    Generates a report
    """

    body = HTMLContainer(
        type="sections",
        container_items = render_report(data=data, report_name=name, minimal=minimal)
    )

    return HTMLBase(body=body, name='Data Report' if name is None else name)



import os
import warnings
from pathlib import Path
from typing import List

import pandas as pd

from xurpas_data_quality.report import get_report, get_empty_report, get_comparison_report, get_error_report, get_test_report
from xurpas_data_quality.data import check_dtypes, describe, load_dataframe,validate_dataframe ,check_col_names, describe_invalid, check_data
from xurpas_data_quality.config import Settings

warnings.filterwarnings("ignore", category=UserWarning, module='visions')

class DataReport:
    def __init__(self, file:str=None, 
                 df:pd.DataFrame|List[pd.DataFrame]|tuple =None, 
                 report_name:str="Data Report", 
                 file_path:str="report.html",
                 data_types:dict=None,
                 minimal:bool = True,
                 config: Settings = Settings,
                 **kwargs):
        
        """
        Initializes the DataReport object
        Args
            file:        The path of the file you want to analyze. If empty, df parameter must exist.
                         Only supports .csv, .xlsx, .parquet, and .orc file formats
            df:          Pandas DataFrame object of data to be analyzed, If using df, file must be empty.
            report_name: Name of the report. Defaults to 'Data Report'
            file_path:   Path/ directory of where the report is to be saved
            data_types:  A dictionary containing the column names and their corresponding data types.
                         If empty, then the data types will be inferred.
            minimal:     Default is True. Check if you want minimal mode as your data report.

        """
        if isinstance(df, list):
            self.df = df
            self.render_empty = False

        elif isinstance(df, tuple):
            self.df = df[0] if isinstance(df[0], pd.DataFrame) else df[0].toPandas()
            self.df_invalid = df[1] if isinstance(df[1], pd.DataFrame) else df[1].toPandas()
            self.errors = df[2] if isinstance(df[2],list) else [df[2]]
            self.empty_df_invalid = True if not validate_dataframe(self.df_invalid) else False
        
        else:
            if file is not None:
                if df is not None:
                    raise KeyError("Only 'file' or 'df' should be used one at a time!")
                self.df = load_dataframe(file)
            else:
                if df is None:
                    raise ValueError("Please provide your data in 'file' or 'df' parameters!")
                
                self.df = df
                self.render_empty = True if not validate_dataframe(df) else False # checks if dataframe is empty

        self.report_name = report_name
        self.minimal = minimal

        if data_types is not None:
            self.data_types = check_dtypes(check_col_names(data_types, df.columns))
        else:
            self.data_types = None

        def has_extension(file_path:str):
            return os.path.splitext(file_path)[1] != '' 
        
        if has_extension(file_path):
            self.save_path = file_path
        else:
            self.save_path = '/'.join([file_path,"report.html"])
            warnings.warn("File name not provided, saving as {file_path}/report.html")

    def describe_dataframe(self):
        self.description = describe(df=self.df, data_types=self.data_types)
        if "cbs_shipment_distribution" in self.report_name:
            self.description.alerts = check_data(self.description.df)

    def describe_invalid_dataframe(self):
        self.description_invalid = describe_invalid(df=self.df_invalid)

    def get_data_quality_report(self, minimal:bool,report_name=None):
        self.describe_dataframe()
        report = get_report(self.description, minimal=minimal,name=report_name)
        return report.render()
    
    def _render_empty_report(self, name=None):
        report = get_empty_report(self.df, name)
        return report.render()
    
    def _render_comparison(self, name=None):
        self.describe_dataframe()
        report = get_comparison_report(self.description, name)
        return report.render()
    
    def _render_error_report(self, name=None):
        self.describe_dataframe()
        self.describe_invalid_dataframe()
        report = get_error_report(data=self.description,
                                  invalid_data=self.description_invalid,
                                  errors=self.errors, 
                                  name = name, is_empty=self.empty_df_invalid, minimal=self.minimal)
        return report.render()
    
    def __render_test_report(self):
        self.describe_dataframe()
        report = get_test_report(df=self.description, name=None)
        rendered_report = report.render()
        return rendered_report

    def to_file_test(self):
        output =Path(self.save_path)
        output.write_text(self.__render_test_report(), encoding='utf-8')

    def to_file(self, report_name:str=None, file_name:str="report.html"):
        output = Path(self.save_path)

        if hasattr(self, 'errors'):
            print(f"saving error report as {self.save_path}")
            output.write_text(self._render_error_report(self.report_name), encoding='utf-8')
            print(f"saved!")

        elif self.render_empty:
            print(f"saving empty report as {self.save_path}")
            output.write_text(self._render_empty_report(self.report_name), encoding='utf-8')
            print("saved!")

        elif isinstance(self.df, list):
            print(f"saving comparison report!")
            output.write_text(self._render_comparison(self.report_name), encoding='utf-8')
            print('saved!')
        
        else:
            print(f"saving as {self.save_path}")
            if self.minimal:
                print("saving minimal version of report!")
                from minify_html import minify
                minified_report = minify(self.get_data_quality_report(self.minimal, self.report_name))
                output.write_text(minified_report, encoding='utf-8')
            else:
                output.write_text(self.get_data_quality_report(self.minimal, self.report_name), encoding='utf-8')
            
            print(f"saved!")



from xurpas_data_quality.data.descriptions import TableDescription
from xurpas_data_quality.render.renderer import HTMLBase
from xurpas_data_quality.render.render_compare import render_compare

def get_comparison_report(data:TableDescription, name:str)-> HTMLBase:
    """
    Generates a comparison report
    """

    return render_compare(data, name)


import pandas as pd
from xurpas_data_quality.render.renderer import HTMLBase, HTMLContainer, HTMLTable, HTMLPlot
from xurpas_data_quality.data.descriptions import TableDescription

def render_error(data:TableDescription, data_invalid:TableDescription, errors:list):
    total_rows = data.dataset_statistics['dataset_length'] + data_invalid.dataset_statistics['dataset_length']
    errors_stats = {
        'Valid Data Ingested (# of Rows)': data.dataset_statistics['dataset_length'],
        'Valid Data Ingested (% Percentage)': "{:0.2f}%".format((data.dataset_statistics['dataset_length']/total_rows)*100),
        'Invalid Data Ingested (# of Rows)': data_invalid.dataset_statistics['dataset_length'],
        'Invalid Data Ingested (% Percentage)': "{:0.2f}%".format((data_invalid.dataset_statistics['dataset_length']/total_rows)*100),
        'Total Rows for Dataset': total_rows
    }
    
    sample_errors_section = HTMLContainer(
        type="default",
        name="Invalid Ingested Data",
        id ="sample-errors",
        container_items=[
            HTMLTable(
                id = "sample",
                data=data_invalid.df.to_html(classes="table table-sm", border=0, index=False, justify='left')
            )
        ]
    )

    errors_list = HTMLContainer(
            type="default",
            name="Errors",
            id = "list-errors",
            container_items=[
                HTMLTable(
                    id = "errors",
                    data = errors
                )
            ]
        )
    
    errors_info = HTMLContainer(
        type="default",
        id = "info-errors",
        name="Error Overview",
        container_items=[
            HTMLContainer(type="default",
                          container_items=[]),
            HTMLTable(
                name="Errors Overview",
                data = errors_stats
            )
        ]
    )

    errors_section = HTMLContainer(
        type="box",
        name="Errors during Ingestion",
        id = "errors_section",
        container_items=[
            HTMLContainer(
                type="tabs",
                container_items=[errors_info,errors_list, sample_errors_section]
            )
        ]
    )

 
    return errors_section



from typing import List

import pandas as pd

def get_compare(df: pd.DataFrame):
    # for every df, get, number of unique values, number of values per table, what values exist in other tables
    
    value_counts = len(df['volume'].value_counts(dropna=True))
    unique_counts = df['volume'].nunique(dropna=True)

    return {
        'df': df,
        'value_count':value_counts,
        'distinct_count':unique_counts
    }

    